#!/bin/python3
"""
Creates a set of waypoints for drone mapping flight plans.

Arguments:
    Required:
        - A bounding box (see docstring on squaregrid function for details),
          should be in EPSG:3857 Pseudo Mercator coordinates (metres)
        - An output filepath
    Optional:
        - A Digital Elevation Model as a GDAL-readable GeoTIFF raster
          (--input_raster). If provided, the program will attempt to sample the
          DEM at each point, providing an elevation as well as x and y coords.

Returns:
    An output CSV file containing the columns:
    [index, x, y] where x and y are coordinates in EPSG:3857
    or, if a DEM is provided, and elevation column will be added:
    [index, x, y, elevation]. Elevation will be in whatever Coordinate
    Reference System that the DEM file is in: be careful!

Typical usage example:

With a DEM (creates points with sampled elevation):
./flightPlanWaypointGenerator.py -minx 12872498.8 -miny -936470.4 -maxx 12873345.9 -maxy -935840.1 -xspac 24 -yspac 36 -ir /path/to/DEM.tif /path/to/outfile.csv

Without a DEM (no elevation column):
./flightPlanWaypointGenerator.py -minx 12872498.8 -miny -936470.4 -maxx 12873345.9 -maxy -935840.1 -xspac 24 -yspac 36 /path/to/outfile.csv


"""

__version__ = "2024-07-31"

import csv
import argparse

# Moving this import to the relevant function because it'll fail if GDAL
# is not installed on the system; only needed if sampling elevation from a DEM
# import sampleRasterAtPoints as sr


def squaregrid(
    minx: float, miny: float, maxx: float, maxy: float, xspac: float, yspac: float
) -> list:
    """
    Create a set of waypoints
    Args: minx, miny, maxx, maxy: coordinates in metres using the
          EPSG:3857 Pseudo Mercator Coordinate Reference System

          xspac, yspac: spacing between forward (xspac) and lateral (yspac)

    Returns: A list, each item of which is a list of points as [index, x, y].
    """
    xlen = maxx - minx
    ylen = maxy - miny
    xpoints = int(xlen / xspac + 1)
    ypoints = int(ylen / yspac + 1)
    points = []
    # loop through lines
    idx = 1
    for yi in range(0, ypoints, 2):
        # points heading out
        for xi in range(xpoints):
            x = minx + (xi * xspac)
            y = miny + (yi * yspac)
            points.append([idx, x, y])
            idx += 1
        # points returning
        for xi in range(xpoints, 0, -1):
            x = minx + ((xi - 1) * xspac)
            y = miny + ((yi + 1) * yspac)
            points.append([idx, x, y])
            idx += 1
    return points


def grid2csv(grid, outfile):
    """Writes a CSV file from a grid without an elevation column"""
    with open(outfile, "w") as of:
        w = csv.writer(of)
        w.writerow(["idx", "x", "y"])
        w.writerows(grid)


def gridWithElevation2csv(grid, outfile):
    """Writes a CSV file from a grid with an elevation column"""
    with open(outfile, "w") as of:
        w = csv.writer(of)
        w.writerow(["idx", "x", "y", "elevation"])
        w.writerows(grid)


if __name__ == "__main__":
    """
    Call the module from the console. See module docstring for usage details.
    """
    p = argparse.ArgumentParser()

    p.add_argument("outfile", help="output csv file")
    p.add_argument("-minx", "--min-x", type=float, help="minimum x coord in EPSG:3857")
    p.add_argument("-maxx", "--max-x", type=float, help="maximum x coord in EPSG:3857")
    p.add_argument("-miny", "--min-y", type=float, help="minimum y coord in EPSG:3857")
    p.add_argument("-maxy", "--max-y", type=float, help="maximum y coord in EPSG:3857")
    p.add_argument("-xspac", "--x-spacing", type=float, help="Forward spacing")
    p.add_argument("-yspac", "--y-spacing", type=float, help="Lateral spacing")
    p.add_argument(
        "-ir", "--input_raster", type=str, help="Digital Elevation Model GeoTIFF file"
    )

    a = p.parse_args()
    print(a)

    # TODO generate the minx, miny, maxx, maxy coords from a bounding box of
    # a supplied Area of Interest (AOI) layer from a GIS file.
    # TODO Arbitrary rotation to any angle (currently only creates waypoints
    # oriented straight east-west
    # TODO clip waypoints by a polygon boundary
    grid = squaregrid(a.min_x, a.min_y, a.max_x, a.max_y, a.x_spacing, a.y_spacing)

    if a.input_raster:
        try:
            import sampleRasterAtPoints as sr

            grid_with_elevation = sr.sampleRasterFromPointsList(a.input_raster, grid)
            gridfile = gridWithElevation2csv(grid_with_elevation, a.outfile)
        except Exception as e:
            print(e)
            print("Maybe your DEM and AOI don't align or are incompatible?")

    else:
        gridfile = grid2csv(grid, a.outfile)



import json
import argparse
import geojson
from typing import Union
from geojson import FeatureCollection


def create_placemarks(
    waypoints_geojson: Union[str, FeatureCollection, dict], parameters: dict
):
    """
    Arguments:
        waypoints_geojson: The waypoint coordinates to be included in the flightplan mission
        parameters: The drone flight parameters in a json
    Returns:
        Geojson object.
    """

    ground_speed = parameters["ground_speed"]
    agl = parameters["altitude_above_ground_level"]

    try:
        first_point = waypoints_geojson["features"][0]
        base_elevation = first_point["geometry"]["coordinates"][2]
    except IndexError:
        base_elevation = 0

    for feature in waypoints_geojson["features"]:
        coords = feature["geometry"]["coordinates"]
        try:
            elevation = coords[2]
            difference_in_elevation = base_elevation - elevation
            altitude = agl - difference_in_elevation
            coords[2] = altitude
        except IndexError:
            altitude = agl
            coords.append(altitude)

        feature["properties"]["speed"] = ground_speed
        feature["properties"]["altitude"] = altitude

    return waypoints_geojson


def main(args_list: list[str] | None = None):
    def json_to_dict(value):
        try:
            return json.loads(value)
        except json.JSONDecodeError:
            raise argparse.ArgumentTypeError(f"Invalid JSON string: {value}")

    parser = argparse.ArgumentParser(
        description="Generate placemark data for drone missions."
    )
    parser.add_argument(
        "--waypoints_geojson",
        required=True,
        type=str,
        help="The waypoint coordinates to be included in the flightplan mission",
    )
    parser.add_argument(
        "--parameters",
        type=json_to_dict,
        help="The drone flight parameters in a json",
    )

    parser.add_argument("--outfile", required=True, help="output GeoJSON file")

    args = parser.parse_args(args_list)

    if "altitude_above_ground_level" not in args.parameters:
        raise ValueError(
            "altitude_above_ground_level is missing in the parameters json"
        )
    if "ground_speed" not in args.parameters:
        raise ValueError("ground_speed is missing in the parameters")

    inpointsfile = open(args.waypoints_geojson, "r")
    points = inpointsfile.read()

    placemark_data = create_placemarks(geojson.loads(points), args.parameters)

    with open(args.outfile, "w") as f:
        f.write(geojson.dumps(placemark_data, indent=2))


if __name__ == "__main__":
    main()

# python3 create_placemarks.py --waypoints_geojson '/home/niraj/NAXA/HOT/drone-flightplan/drone_flightplan/waypoints_2.geojson' --parameters '{"forward_photo_height": 84.0, "side_photo_width": 149.0, "forward_spacing": 20.95, "side_spacing": 44.6, "ground_speed": 10.47, "altitude_above_ground_level": 118}'



"""Drone FlightPlan Package."""

import os
from drone_flightplan.calculate_parameters import calculate_parameters
from drone_flightplan.create_placemarks import create_placemarks
from drone_flightplan.waypoints import create_waypoint
from drone_flightplan.add_elevation_from_dem import add_elevation_from_dem
from drone_flightplan.wpml import create_wpml
from drone_flightplan.create_flightplan import create_flightplan

package_root = os.path.dirname(os.path.abspath(__file__))



#!/usr/bin/python3
"""
Sample a Digital Elevation Model raster values at points. Primarily intended
as a library; other Python modules can use it to add an elevation attribute
to a file of autogenerated waypoints.

However, it can be used standalone to sample a given DEM (GeoTIFF) at the
points given by a given GeoJSON point layer.

Used from the command line, it expects the points to be a GeoJSON file on the
local filesystem.
"""

import argparse
from osgeo import ogr, gdal, osr
import math
import struct


def raster_data_format_string(input_datatype: str):
    """
    Returns a format string for unpacking a c-style Struct object based
    on the appropriate GDAL raster band data type. Probably not correct
    for all data types; at the moment we're only concerned with the ones
    likely to be encountered in Digital Elevation Models.
    """
    type_map = {
        "Byte": "b",
        "Int8": "b",
        "UInt16": "<H",
        "Int16": "<h",
        "UInt32": "L",
        "Int32": "l",
        "UInt64": "q",
        "Int64": "Q",
        "Float32": "f",
        "Float64": "d",
    }
    struct_data_type = type_map[input_datatype]
    return struct_data_type


def add_elevation_from_dem(raster_file, points, outfile):
    """
    Arguments:
        DEM raster file as GeoTIFF
        Points as GeoJSON string
    Returns:
        Writes GeoJSON file with added elevation attribute on each point
    """
    r = gdal.Open(raster_file)

    # Create an empty Spatial Reference object
    # and set it to the CRS of the input raster
    rasterSR = osr.SpatialReference()
    rasterSR.ImportFromProj4(r.GetProjection())
    print(f"\nRaster Coordinate Reference System: \n{rasterSR}")

    # Get the raster band (if it's a DEM, this should be the only band)
    # TODO this would be a good time to check that it's a single-band
    # raster with values that make sense for a DEM
    band = r.GetRasterBand(1)

    # Determine the data type
    raster_data_type = gdal.GetDataTypeName(band.DataType)
    print(f"\nRaster band 1 data type: {raster_data_type}")
    struct_data_type = raster_data_format_string(raster_data_type)
    print(
        f"The GDAL data type is {raster_data_type}, which hopefully "
        f"corresponds to a struct {struct_data_type} data type"
    )

    # Create the tranforms between geographical and pixel coordinates
    # The forward transform takes pixel coords and returns geographical coords,
    # the reverse transform... well, you get the idea.
    forward = r.GetGeoTransform()
    reverse = gdal.InvGeoTransform(forward)

    p = ogr.Open(points)
    lyr = p.GetLayer()
    pointSR = lyr.GetSpatialRef()
    pointLD = lyr.GetLayerDefn()
    print(f"\nPoint layer Coordinate Reference System: {pointSR.GetName()}")

    transform = osr.CoordinateTransformation(pointSR, rasterSR)

    # Create the GDAL GeoJSON output file infrastructure
    outDriver = ogr.GetDriverByName("GeoJSON")
    outDataSource = outDriver.CreateDataSource(outfile)
    outLayer = outDataSource.CreateLayer("waypoints", pointSR, ogr.wkbPoint)

    # Add an elevation field for easy reference (the point is XYZ but
    # it's nice to have access to the elevation as a property)
    elevation_field = ogr.FieldDefn("elevation", ogr.OFTReal)
    outLayer.CreateField(elevation_field)
    # Add fields from input layer to output layer
    fields = []
    for i in range(pointLD.GetFieldCount()):
        fd = pointLD.GetFieldDefn(i)
        fields.append(fd)
    for fd in fields:
        print(f"Adding field {fd.name} of type {fd.GetTypeName()}.")
        outLayer.CreateField(fd)
    featureDefn = outLayer.GetLayerDefn()

    for feature in lyr:
        geom = feature.GetGeometryRef()
        pointXYRasterCRS = transform.TransformPoint(geom.GetX(), geom.GetY())
        mapX = pointXYRasterCRS[1]
        mapY = pointXYRasterCRS[0]
        pixcoords = gdal.ApplyGeoTransform(reverse, mapX, mapY)
        pixX = math.floor(pixcoords[0])
        pixY = math.floor(pixcoords[1])
        elevation = 0
        try:
            elevationstruct = band.ReadRaster(pixX, pixY, 1, 1)
            ele = struct.unpack(struct_data_type, elevationstruct)[0]
            elevation = round(ele, 1)
            if elevation == -9999 or elevation == -9999.0:
                # It's almost certainly a nodata value, possibly over water
                elevation = 0
        except Exception:
            # print('Something went wrong with the raster sampling'
            #      f' at point {geom.GetX()}, {geom.GetY()}')
            # print(e)
            pass
        new_point = ogr.Geometry(ogr.wkbPoint)
        new_point.AddPoint(geom.GetX(), geom.GetY(), elevation)
        outFeature = ogr.Feature(featureDefn)
        outFeature.SetGeometry(new_point)
        outFeature.SetField("elevation", elevation)
        for fd in fields:
            val = feature.GetField(fd.name)
            outFeature.SetField(fd.name, val)
        outLayer.CreateFeature(outFeature)
    return 0


if __name__ == "__main__":
    p = argparse.ArgumentParser()

    p.add_argument("inraster", help="input DEM GeoTIFF raster file")
    p.add_argument("inpoints", help="input points geojson file")
    p.add_argument("outfile", help="output GeoJSON file")

    a = p.parse_args()

    inpointsfile = open(a.inpoints, "r")
    points = inpointsfile.read()

    writeout = add_elevation_from_dem(a.inraster, points, a.outfile)



#!/usr/bin/python3
"""
Sample a Digital Elevation Model raster values at points. Primarily intended
as a library; the associated flightPlanWaypointGenerator uses it to add an
elevation column to a grid of autogenerated waypoints.

However, it can be used standalone to sample a given DEM (GeoTIFF) at the
points given by a given GeoJSON point layer.
"""

import argparse

from osgeo import ogr, gdal, osr
import math
import struct
import csv


def sampleRasterFromPointsList(raster_file, points_list):
    """
    Arguments:
    DEM raster file as geotiff
    A list of points. Each point consists of [index, x, y] in EPSG:3857
    """

    r = gdal.Open(raster_file)
    rasterSR = osr.SpatialReference()
    rasterSR.ImportFromProj4(r.GetProjection())
    print("\nRaster Coordinate Reference System:")
    print(rasterSR)
    band = r.GetRasterBand(1)
    print("\nRaster band 1 data type:")
    rasterDataType = gdal.GetDataTypeName(band.DataType)
    print(rasterDataType)
    forward = r.GetGeoTransform()
    reverse = gdal.InvGeoTransform(forward)
    print(f"\nRaster forward geotransform: {forward}")
    print(f"\nRaster reverse geotransform: {reverse}")

    pointSR = osr.SpatialReference()
    pointSR.ImportFromEPSG(3857)
    # print(pointSR.GetName)

    transform = osr.CoordinateTransformation(pointSR, rasterSR)
    # print(f'\nPoint layer transform: {transform}')

    points = []

    for feature in points_list:
        pointXYRasterCRS = transform.TransformPoint(feature[1], feature[2])
        mapX = pointXYRasterCRS[1]
        mapY = pointXYRasterCRS[0]
        pixcoords = gdal.ApplyGeoTransform(reverse, mapX, mapY)
        pixX = math.floor(pixcoords[0])
        pixY = math.floor(pixcoords[1])
        elevationstruct = band.ReadRaster(pixX, pixY, 1, 1)
        # TODO The struct needs to be unpacked based on the data type in the DEM.
        # Currently we're unpacking a littlendian int16 ('<h'), which
        # corresponds to what we're getting from the global JAXA DEM, but that
        # won't be consistent across DEMs.
        elevation = struct.unpack("<h", elevationstruct)[0]
        # print(f'Point coordinates in point layer CRS: {geom}')
        # print(f'Point coordinates in raster layer CRS: {mapX}, {mapY}')
        # print(f'Pixel coordinates: {pixX}, {pixY}')
        # print(f'Elevation at point: {elevation}')
        points.append([feature[0], feature[1], feature[2], elevation])
    return points


def rasterValuesAtPoints(raster_file, point_file):
    """
    Arguments:
    DEM raster file as geotiff
    Point file as GeoJSON
    """

    r = gdal.Open(raster_file)
    rasterSR = osr.SpatialReference()
    # rasterSR.ImportFromEPSG(4326) #ONLY FOR TESTING
    rasterSR.ImportFromProj4(r.GetProjection())
    print("\nRaster Coordinate Reference System:")
    print(rasterSR)
    band = r.GetRasterBand(1)
    print("\nRaster band 1 data type:")
    rasterDataType = gdal.GetDataTypeName(band.DataType)
    print(rasterDataType)
    forward = r.GetGeoTransform()
    reverse = gdal.InvGeoTransform(forward)
    print(f"\nRaster forward geotransform: {forward}")
    print(f"\nRaster reverse geotransform: {reverse}")

    p = ogr.Open(point_file)
    lyr = p.GetLayer()
    print("\nPoint layer Coordinate Reference System:")
    pointSR = lyr.GetSpatialRef()
    print(pointSR.GetName)

    transform = osr.CoordinateTransformation(pointSR, rasterSR)
    print(f"\nPoint layer transform: {transform}")
    points = []
    for feature in lyr:
        geom = feature.GetGeometryRef()
        pointXYRasterCRS = transform.TransformPoint(geom.GetX(), geom.GetY())
        mapX = pointXYRasterCRS[1]
        mapY = pointXYRasterCRS[0]
        pixcoords = gdal.ApplyGeoTransform(reverse, mapX, mapY)
        pixX = math.floor(pixcoords[0])
        pixY = math.floor(pixcoords[1])
        elevationstruct = band.ReadRaster(pixX, pixY, 1, 1)
        elevation = struct.unpack("f", elevationstruct)[0]
        # print(f'Point coordinates in point layer CRS: {geom}')
        # print(f'Point coordinates in raster layer CRS: {mapX}, {mapY}')
        # print(f'Pixel coordinates: {pixX}, {pixY}')
        # print(f'Elevation at point: {elevation}')

        # TODO: check what CRS this returns, probably x and y should
        # be returned in the same CRS as the input file

        # TODO: this won't include an index, and will discard any fields
        # that came with the original GeoJSON.
        points.append([mapX, mapY, elevation])
    return points


def gridWithElevation2csv(grid, outfile):
    """Writes a CSV file from a grid with an elevation column"""
    with open(outfile, "w") as of:
        w = csv.writer(of)
        w.writerow(["x", "y", "elevation"])
        w.writerows(grid)


if __name__ == "__main__":
    p = argparse.ArgumentParser()

    p.add_argument("inraster", help="input DEM GeoTIFF raster file")
    p.add_argument("inpoints", help="input points geojson file")
    p.add_argument("outfile", help="output csv file")

    a = p.parse_args()

    grid = rasterValuesAtPoints(a.inraster, a.inpoints)
    gridfile = gridWithElevation2csv(grid, a.outfile)



import argparse
import geojson
import logging
from typing import Union
from geojson import FeatureCollection
from drone_flightplan.calculate_parameters import calculate_parameters
from drone_flightplan.waypoints import create_waypoint
from drone_flightplan.add_elevation_from_dem import add_elevation_from_dem
from drone_flightplan.create_placemarks import create_placemarks
from drone_flightplan.wpml import create_wpml

# Instantiate logger
log = logging.getLogger(__name__)


def create_flightplan(
    aoi: Union[str, FeatureCollection, dict],
    forward_overlap: float,
    side_overlap: float,
    agl: float,
    gsd: float = None,
    image_interval: int = 2,
    dem: str = None,
    outfile: str = None,
    generate_each_points: bool = False,
    rotation_angle: float = 0.0,
    take_off_point: list[float] = None,
):
    """
    Arguments:
        aoi: The area of interest in GeoJSON format.
        forward_overlap: The forward overlap in percentage.
        side_overlap: The side overlap in percentage.
        agl: The altitude above ground level in meters.
        gsd: The ground sampling distance in cm/px.
        image_interval: The time interval between two consecutive images in seconds.
        rotation_angle: The rotation angle for the flight grid in degrees.
    Returns:
        Drone Flightplan in kmz format
    """

    parameters = calculate_parameters(
        forward_overlap, side_overlap, agl, gsd, image_interval
    )

    waypoints = create_waypoint(
        aoi,
        agl,
        gsd,
        forward_overlap,
        side_overlap,
        rotation_angle,
        generate_each_points,
        take_off_point=take_off_point,
    )

    # Add elevation data to the waypoints
    if dem:
        # TODO: Do this with inmemory data
        outfile_with_elevation = "/tmp/output_file_with_elevation.geojson"
        add_elevation_from_dem(dem, waypoints, outfile_with_elevation)

        inpointsfile = open(outfile_with_elevation, "r")
        waypoints = inpointsfile.read()

    # calculate the placemark data
    placemarks = create_placemarks(geojson.loads(waypoints), parameters)

    # create wpml file
    outpath = create_wpml(placemarks, outfile)
    log.info(f"Flight plan generated in the path {outpath}")
    return outpath


def validate_coordinates(value):
    try:
        lon, lat = map(float, value.split(","))
        if not (-180 <= lon <= 180 and -90 <= lat <= 90):
            raise argparse.ArgumentTypeError(
                "Coordinates must be in the format 'longitude,latitude' and within valid ranges."
            )
        return [lon, lat]
    except ValueError:
        raise argparse.ArgumentTypeError(
            "Invalid format. Coordinates must be in 'longitude,latitude' format."
        )


def main():
    parser = argparse.ArgumentParser(
        description="Generate waypoints for drone missions."
    )
    parser.add_argument(
        "--project_geojson",
        required=True,
        type=str,
        help="The GeoJSON polygon representing the area of interest.",
    )

    group = parser.add_mutually_exclusive_group(required=True)

    group.add_argument(
        "--altitude_above_ground_level",
        type=float,
        help="The flight altitude in meters.",
    )
    group.add_argument(
        "--gsd",
        type=float,
        help="The ground sampling distance in cm/px.",
    )

    parser.add_argument(
        "--forward_overlap",
        type=float,
        default=70.0,
        help="The forward overlap in percentage.",
    )
    parser.add_argument(
        "--side_overlap",
        type=float,
        default=70.0,
        help="The side overlap in percentage.",
    )
    parser.add_argument(
        "--image_interval",
        type=int,
        default=2,
        help="The time interval between two consecutive images in seconds.",
    )

    parser.add_argument("--inraster", help="input DEM GeoTIFF raster file")
    parser.add_argument("--outfile", required=True, help="output GeoJSON file")
    parser.add_argument(
        "--generate_each_points",
        action="store_true",
        help="Do you want waypoints or waylines.",
    )
    parser.add_argument(
        "--rotation_angle",
        type=float,
        default=0.0,
        help="The rotation angle for the flight grid in degrees.",
    )

    parser.add_argument(
        "--take_off_point",
        required=True,
        type=validate_coordinates,
        help="Take off Point Coordinates in 'longitude,latitude' format (e.g., 82.52,28.29).",
    )
    args = parser.parse_args()

    with open(args.project_geojson, "r") as f:
        aoi = geojson.load(f)

    create_flightplan(
        aoi,
        args.forward_overlap,
        args.side_overlap,
        args.altitude_above_ground_level,
        args.gsd,
        args.image_interval,
        args.inraster,
        args.outfile,
        args.generate_each_points,
        args.rotation_angle,
        args.take_off_point,
    )


if __name__ == "__main__":
    main()


# python3 create_flightplan.py --project_geojson '/home/niraj/NAXA/HOT/adarsha_polygons_for_terrain_testing.geojson'  --altitude_above_ground_level 118 --forward_overlap 75 --side_overlap 70 --image_interval 2 --inraster '/home/niraj/Downloads/Bhanu.tif'  --outfile /home/niraj/NAXA/HOT/drone-flightplan/drone_flightplan --generate_each_points



import logging
import argparse
from math import sqrt
import pyproj
import geojson
from shapely.geometry import Point, shape, Polygon
from shapely.affinity import rotate
from shapely.ops import transform
from drone_flightplan.calculate_parameters import calculate_parameters as cp


log = logging.getLogger(__name__)


def generate_grid_in_aoi(
    aoi_polygon: shape, x_spacing: float, y_spacing: float
) -> list[Point]:
    """
    Generate a grid of points within a given Area of Interest (AOI) polygon.

    Parameters:
        aoi_polygon (shape): The Shapely shape representing the area of interest.
        x_spacing (float): The spacing between points along the x-axis (in meters).
        y_spacing (float): The spacing between points along the y-axis (in meters).

    Returns:
        list[Point]: A list of Points representing the generated grid within the AOI.
    """
    minx, miny, maxx, maxy = aoi_polygon.bounds
    xpoints = int((maxx - minx) / x_spacing) + 1
    ypoints = int((maxy - miny) / y_spacing) + 1

    points = []
    for yi in range(ypoints):
        for xi in range(xpoints):
            x = minx + xi * x_spacing
            y = miny + yi * y_spacing
            point = Point(x, y)
            if aoi_polygon.contains(point):
                points.append(point)

    return points


def calculate_distance(point1, point2):
    """
    Calculate Euclidean distance between two points.
    """
    return sqrt((point1.x - point2.x) ** 2 + (point1.y - point2.y) ** 2)


def create_path(
    points: list[Point],
    forward_spacing: float,
    generate_3d: bool = False,
    take_off_point: list[float] = None,
) -> list[dict]:
    """
    Create a continuous path of waypoints from a grid of points.

    Parameters:
        points (list[Point]): A list of Points representing the grid.
        forward_spacing (float): The spacing between rows of points (in meters).
        generate_3d (bool): Whether to generate additional 3D waypoints for the path.

    Returns:
        list[dict]: A list of dictionaries representing the waypoints along the path.
    """
    rows = {}
    for point in points:
        row_key = round(point.y, 8)
        if row_key not in rows:
            rows[row_key] = []
        rows[row_key].append(point)

    continuous_path = []
    for idx, row in enumerate(sorted(rows.keys())):
        row_points = sorted(rows[row], key=lambda p: p.x)
        if idx % 2 == 1:
            row_points.reverse()

        # initialize points at the start and end of each row
        first_point = row_points[0]
        last_point = row_points[-1]

        # define coordinates for extra points
        start_extra_point = Point(
            first_point.x - (forward_spacing if idx % 2 == 0 else -forward_spacing),
            first_point.y,
        )
        end_extra_point = Point(
            last_point.x + (forward_spacing if idx % 2 == 0 else -forward_spacing),
            last_point.y,
        )

        # Add the extra points with no photo taken
        continuous_path.append(
            {
                "coordinates": start_extra_point,
                "angle": -90 if idx % 2 == 0 else 90,
                "take_photo": False,
                "gimbal_angle": "-90",
            }
        )

        # Add each point with its associated properties
        for point in row_points:
            continuous_path.append(
                {
                    "coordinates": point,
                    "angle": -90 if idx % 2 == 0 else 90,
                    "take_photo": True,
                    "gimbal_angle": "-90",
                }
            )

        # Add the extra point at the end with no photo taken
        continuous_path.append(
            {
                "coordinates": end_extra_point,
                "angle": -90 if idx % 2 == 0 else 90,
                "take_photo": False,
                "gimbal_angle": "-90",
            }
        )

        if generate_3d:
            continuous_path.extend(
                generate_3d_waypoints(
                    row_points, idx, angle=(-90 if idx % 2 == 0 else 90)
                )
            )

    return continuous_path


def generate_3d_waypoints(
    row_points: list[Point], row_index: int, angle: int
) -> list[dict]:
    """
    Generate additional 3D waypoints by alternating the gimbal angle for each row.

    Parameters:
        row_points (list[Point]): A list of Points in the current row.
        row_index (int): The index of the current row.
        angle (int): The angle at which the gimbal should be tilted.

    Returns:
        list[dict]: A list of dictionaries representing the additional 3D waypoints.
    """
    # Return path with -45 degree angle
    return_path = [
        {
            "coordinates": wp,
            "angle": str(-angle),
            "take_photo": True,
            "gimbal_angle": "-45"
            if row_index % 2 == 0
            else "45",  # Alternate angles based on row index
        }
        for wp in reversed(row_points)
    ]
    return_path[0]["take_photo"] = False
    return_path[-1]["take_photo"] = False

    # Forward path with 45 degree angle
    forward_path = [
        {
            "coordinates": wp,
            "angle": str(angle),
            "take_photo": True,
            "gimbal_angle": "45"
            if row_index % 2 == 0
            else "-45",  # Alternate angles based on row index
        }
        for wp in row_points
    ]
    forward_path[0]["take_photo"] = False
    forward_path[-1]["take_photo"] = False

    return return_path + forward_path


def exclude_no_fly_zones(points: list[dict], no_fly_zones: list[Polygon]) -> list[dict]:
    """
    Exclude waypoints that fall within defined no-fly zones.

    Parameters:
        points (list[dict]): A list of waypoints.
        no_fly_zones (list[Polygon]): A list of Polygons representing no-fly zones.

    Returns:
        list[dict]: A list of waypoints excluding those within no-fly zones.
    """
    return [
        point
        for point in points
        if not any(nfz.contains(point["coordinates"]) for nfz in no_fly_zones)
    ]


def remove_middle_points(data):
    processed_data = []
    i = 0

    while i < len(data):
        current_angle = data[i]["angle"]
        segment_start = i

        # Find the end of the segment with the same angle
        while i < len(data) and data[i]["angle"] == current_angle:
            i += 1

        segment_end = i

        # If the segment has more than 4 points, keep only the first 2 and the last 2
        if segment_end - segment_start > 4:
            processed_data.extend(data[segment_start : segment_start + 2])
            processed_data.extend(data[segment_end - 2 : segment_end])
        else:
            processed_data.extend(data[segment_start:segment_end])

    # Make take_photo = False for all the points
    for point in processed_data:
        point["take_photo"] = False

    return processed_data


def create_waypoint(
    project_area: dict,
    agl: float,
    gsd: float,
    forward_overlap: float,
    side_overlap: float,
    rotation_angle: float = 0.0,
    generate_each_points: bool = False,
    generate_3d: bool = False,
    no_fly_zones: dict = None,
    take_off_point: list[float] = None,
) -> str:
    """
    Create waypoints for a given project area based on specified parameters.

    Parameters:
        project_area (dict): GeoJSON dictionary representing the project area.
        agl (float): Altitude above ground level.
        gsd (float): Ground Sampling Distance.
        forward_overlap (float): Forward overlap percentage for the waypoints.
        side_overlap (float): Side overlap percentage for the waypoints.
        rotation_angle (float): The rotation angle for the flight grid in degrees.
        generate_each_points (bool): Flag True to generate individual waypoints, False for waylines.
        generate_3d (bool): Flag to determine if 3D waypoints should be generated.
        no_fly_zones (dict, optional): GeoJSON dictionary representing no-fly zones.
    Returns:
        geojson: waypoints generated within the project area in the geojson format

    Example Response:
    {
        "type": "FeatureCollection",
            "features": [
                {
                "type": "Feature",
                "geometry": {
                    "type": "Point",
                    "coordinates": [
                    85.328347,
                    27.729837
                    ]
                },
                "properties": {
                    "index": 0,
                    "angle": "0",
                    "take_photo": false,
                    "gimbal_angle": "-90"
                }
            }
        }
    }

    """
    parameters = cp(forward_overlap, side_overlap, agl, gsd)
    side_spacing = parameters["side_spacing"]
    forward_spacing = parameters["forward_spacing"]

    polygon = shape(project_area["features"][0]["geometry"])

    wgs84 = pyproj.CRS("EPSG:4326")
    web_mercator = pyproj.CRS("EPSG:3857")

    transformer_to_3857 = pyproj.Transformer.from_crs(
        wgs84, web_mercator, always_xy=True
    ).transform
    transformer_to_4326 = pyproj.Transformer.from_crs(
        web_mercator, wgs84, always_xy=True
    ).transform

    polygon_3857 = transform(transformer_to_3857, polygon)

    # Calculate the centroid for centering the grid
    centroid = polygon_3857.centroid

    # Rotate the polygon to the specified angle around the centroid
    rotated_polygon = rotate(
        polygon_3857, rotation_angle, origin=centroid, use_radians=False
    )

    # Generate grid within the rotated AOI
    grid = generate_grid_in_aoi(rotated_polygon, forward_spacing, side_spacing)

    # Create path (either waypoints or waylines) and rotate back to original angle
    initial_path = create_path(grid, forward_spacing, generate_3d=generate_3d)

    # Path initialization
    path = []

    # Conditionally add takeoff point if available
    if take_off_point:

        # Get the first and last point of the initial path
        first_path_point = initial_path[0]["coordinates"]
        last_path_point = initial_path[-1]["coordinates"]

        # Calculate distances from the takeoff point
        distance_to_first = calculate_distance(Point(transformer_to_3857(*take_off_point)), first_path_point)
        distance_to_last = calculate_distance(Point(transformer_to_3857(*take_off_point)), last_path_point)
        if distance_to_last < distance_to_first:
            initial_path.reverse()

        initial_point = {
            "coordinates": Point(transformer_to_3857(*take_off_point)),
            "take_photo": False,
            "angle": 0,
            "gimbal_angle": "-90",
        }
        path.append(initial_point)

    # Add the rest of the points with rotation
    path.extend(
        [
            {
                "coordinates": rotate(
                    point["coordinates"],
                    -rotation_angle,
                    origin=centroid,
                    use_radians=False,
                ),
                "angle": point["angle"],
                "take_photo": point["take_photo"],
                "gimbal_angle": point["gimbal_angle"],
            }
            for point in initial_path
        ]
    )

    # If generating waylines, just add two points at each end
    waypoints = remove_middle_points(path) if not generate_each_points else path

    # If no-fly zones are provided, exclude points that fall inside no-fly zones
    if no_fly_zones:
        no_fly_polygons = [
            transform(transformer_to_3857, shape(zone["geometry"]))
            for zone in no_fly_zones["features"]
        ]
        waypoints = exclude_no_fly_zones(waypoints, no_fly_polygons)

    # Generate GeoJSON features
    features = []
    for index, wp in enumerate(waypoints):
        coordinates_4326 = transformer_to_4326(wp["coordinates"].x, wp["coordinates"].y)
        feature = geojson.Feature(
            geometry=geojson.Point(coordinates_4326),
            properties={
                "index": index,
                "heading": wp["angle"],
                "take_photo": wp["take_photo"],
                "gimbal_angle": wp["gimbal_angle"],
            },
        )
        features.append(feature)
    feature_collection = geojson.FeatureCollection(features)
    return geojson.dumps(feature_collection, indent=2)


def validate_coordinates(value):
    try:
        lon, lat = map(float, value.split(","))
        if not (-180 <= lon <= 180 and -90 <= lat <= 90):
            raise argparse.ArgumentTypeError(
                "Coordinates must be in the format 'longitude,latitude' and within valid ranges."
            )
        return [lon, lat]
    except ValueError:
        raise argparse.ArgumentTypeError(
            "Invalid format. Coordinates must be in 'longitude,latitude' format."
        )


def main():
    """
    The main entry point of the script. Parses command-line arguments and
    generates waypoints for a drone mission based on the provided parameters.
    """
    parser = argparse.ArgumentParser(
        description="Generate waypoints for drone missions."
    )
    parser.add_argument(
        "--project_geojson_polygon",
        required=True,
        type=str,
        help="The GeoJSON polygon representing the area of interest.",
    )
    parser.add_argument(
        "--altitude_above_ground_level",
        required=True,
        type=float,
        help="The flight altitude in meters.",
    )
    parser.add_argument(
        "--forward_overlap",
        type=float,
        default=70.0,
        help="The forward overlap in percentage.",
    )
    parser.add_argument(
        "--side_overlap",
        type=float,
        default=70.0,
        help="The side overlap in percentage.",
    )
    parser.add_argument(
        "--rotation_angle",
        type=float,
        default=0.0,
        help="The rotation angle for the flight grid in degrees.",
    )
    parser.add_argument(
        "--generate_each_points",
        action="store_true",
        help="Do you want waypoints or waylines.",
    )
    parser.add_argument(
        "--generate_3d", action="store_true", help="Generate 3D imagery."
    )
    parser.add_argument(
        "--no_fly_zones", type=str, help="GeoJSON file containing no-fly zones."
    )
    parser.add_argument(
        "--output_file_path",
        type=str,
        required=True,
        help="The output GeoJSON file path for the waypoints.",
    )
    parser.add_argument(
        "--take_off_point",
        required=True,
        type=validate_coordinates,
        help="Take off Point Coordinates in 'longitude,latitude' format (e.g., 82.52,28.29).",
    )
    args = parser.parse_args()

    with open(args.project_geojson_polygon, "r") as f:
        boundary = geojson.load(f)

    no_fly_zones = None
    if args.no_fly_zones:
        with open(args.no_fly_zones, "r") as f:
            no_fly_zones = geojson.load(f)

    coordinates = create_waypoint(
        boundary,
        args.altitude_above_ground_level,
        None,  # GSD can be None if you calculate based on altitude
        args.forward_overlap,
        args.side_overlap,
        args.rotation_angle,
        args.generate_each_points,
        args.generate_3d,
        no_fly_zones,
        args.take_off_point,
    )

    with open(args.output_file_path, "w") as f:
        f.write(coordinates)

    return coordinates


if __name__ == "__main__":
    main()



#!/usr/bin/python3
"""
Temporary utility to test the addElevationFromDEM module. It's a bit of a hassle reading and writing GeoJSON as in-memory objects using GDAL, so testing that from here.

I couldn't be bothered to create a proper unit test.
"""

import argparse
from addElevationFromDEM import add_elevation_from_dem as ae

if __name__ == "__main__":
    p = argparse.ArgumentParser()

    p.add_argument("inraster", help="input DEM GeoTIFF raster file")
    p.add_argument("inpoints", help="input points geojson file")
    p.add_argument("outfile", help="output GeoJSON file")

    a = p.parse_args()

    infile = open(a.inpoints, "r")
    inpointstring = infile.read()

    waypoints_with_elevation = ae(a.inraster, inpointstring)
    print(waypoints_with_elevation)



import argparse
import os
import geojson
import zipfile
import logging
from typing import Union
from geojson import FeatureCollection
import xml.etree.ElementTree as ET
from xml.etree.ElementTree import Element


# Instantiate logger
log = logging.getLogger(__name__)


def zip_directory(directory_path, zip_path):
    with zipfile.ZipFile(zip_path, "w") as zipf:
        for root, dirs, files in os.walk(directory_path):
            for file in files:
                zipf.write(
                    os.path.join(root, file),
                    os.path.relpath(
                        os.path.join(root, file), os.path.join(directory_path, "..")
                    ),
                )


def create_zip_file(waylines_path_uid):
    # Create the wpmz folder if it doesn't exist
    wpmz_path = f"{waylines_path_uid}/wpmz"
    os.makedirs(wpmz_path, exist_ok=True)

    # TODO: Need to check if it is really required. It might not be needed.

    # import xml.etree.ElementTree as ET

    # xml_string = """<?xml version="1.0" encoding="UTF-8"?>
    # <kml xmlns="http://www.opengis.net/kml/2.2" xmlns:wpml="http://www.dji.com/wpmz/1.0.2">
    #   <Document>
    #     <wpml:author>fly</wpml:author>
    #     <wpml:createTime>1702051864938</wpml:createTime>
    #     <wpml:updateTime>1702051864938</wpml:updateTime>
    #     <wpml:missionConfig>
    #       <wpml:flyToWaylineMode>safely</wpml:flyToWaylineMode>
    #       <wpml:finishAction>noAction</wpml:finishAction>
    #       <wpml:exitOnRCLost>executeLostAction</wpml:exitOnRCLost>
    #       <wpml:executeRCLostAction>hover</wpml:executeRCLostAction>
    #       <wpml:globalTransitionalSpeed>2.5</wpml:globalTransitionalSpeed>
    #       <wpml:droneInfo>
    #         <wpml:droneEnumValue>68</wpml:droneEnumValue>
    #         <wpml:droneSubEnumValue>0</wpml:droneSubEnumValue>
    #       </wpml:droneInfo>
    #     </wpml:missionConfig>
    #   </Document>
    # </kml>
    # """

    # # Parse the XML string
    # root = ET.fromstring(xml_string)

    # # Create an ElementTree object
    # tree = ET.ElementTree(root)

    # # Write the ElementTree object to a file
    # with open(f"{wpmz_path}/template.kml", "wb") as file:
    #     tree.write(file, encoding="utf-8", xml_declaration=True)

    # Read content of template.kml
    with open(f"{waylines_path_uid}/waylines.wpml", "r") as f:
        wpml_content = f.read()

    with open(f"{wpmz_path}/waylines.wpml", "w") as f:
        f.write(wpml_content)

    # Create a Zip file containing the contents of the wpmz folder directly
    output_file_name = f"{waylines_path_uid}/output.kmz"
    zip_directory(wpmz_path, output_file_name)

    return output_file_name


def take_photo_action(action_group_element: Element, index: str):
    action = ET.SubElement(action_group_element, "wpml:action")
    action_id = ET.SubElement(action, "wpml:actionId")
    action_id.text = str(index)
    action_actuator_func = ET.SubElement(action, "wpml:actionActuatorFunc")
    action_actuator_func.text = "takePhoto"
    action_actuator_func_param = ET.SubElement(action, "wpml:actionActuatorFuncParam")
    payload_position_index = ET.SubElement(
        action_actuator_func_param, "wpml:payloadPositionIndex"
    )
    payload_position_index.text = "0"


def gimble_rotate_action(action_group_element: Element, index: str, gimbal_angle: str):
    action1 = ET.SubElement(action_group_element, "wpml:action")
    action1_id = ET.SubElement(action1, "wpml:actionId")
    action1_id.text = str(index)
    action1_actuator_func = ET.SubElement(action1, "wpml:actionActuatorFunc")
    action1_actuator_func.text = "gimbalRotate"
    action1_actuator_func_param = ET.SubElement(action1, "wpml:actionActuatorFuncParam")
    gimbal_heading_yaw_base = ET.SubElement(
        action1_actuator_func_param, "wpml:gimbalHeadingYawBase"
    )
    gimbal_heading_yaw_base.text = "aircraft"
    gimbal_rotate_mode = ET.SubElement(
        action1_actuator_func_param, "wpml:gimbalRotateMode"
    )
    gimbal_rotate_mode.text = "absoluteAngle"
    gimbal_pitch_rotate_enable = ET.SubElement(
        action1_actuator_func_param, "wpml:gimbalPitchRotateEnable"
    )
    gimbal_pitch_rotate_enable.text = "1"
    gimbal_pitch_rotate_angle = ET.SubElement(
        action1_actuator_func_param, "wpml:gimbalPitchRotateAngle"
    )
    gimbal_roll_rotate_enable = ET.SubElement(
        action1_actuator_func_param, "wpml:gimbalRollRotateEnable"
    )
    gimbal_roll_rotate_angle = ET.SubElement(
        action1_actuator_func_param, "wpml:gimbalRollRotateAngle"
    )

    # If the gimbal pitch rotate angle is 45, make it -90. 45 is just for reference. We need to change roll angle for this.
    if str(gimbal_angle) == "45":
        gimbal_pitch_rotate_angle.text = "-90"
        gimbal_roll_rotate_enable.text = "1"
        gimbal_roll_rotate_angle.text = "-45"

    else:
        gimbal_pitch_rotate_angle.text = str(gimbal_angle)
        gimbal_roll_rotate_enable.text = "0"
        gimbal_roll_rotate_angle.text = "0"

    gimbal_yaw_rotate_enable = ET.SubElement(
        action1_actuator_func_param, "wpml:gimbalYawRotateEnable"
    )
    gimbal_yaw_rotate_enable.text = "0"
    gimbal_yaw_rotate_angle = ET.SubElement(
        action1_actuator_func_param, "wpml:gimbalYawRotateAngle"
    )
    gimbal_yaw_rotate_angle.text = "0"
    gimbal_rotate_time_enable = ET.SubElement(
        action1_actuator_func_param, "wpml:gimbalRotateTimeEnable"
    )
    gimbal_rotate_time_enable.text = "0"
    gimbal_rotate_time = ET.SubElement(
        action1_actuator_func_param, "wpml:gimbalRotateTime"
    )
    gimbal_rotate_time.text = "0"
    payload_position_index = ET.SubElement(
        action1_actuator_func_param, "wpml:payloadPositionIndex"
    )
    payload_position_index.text = "0"


def create_placemark(placemark):
    try:
        index = placemark["properties"]["index"]
        coordinate = placemark["geometry"]["coordinates"]
        coordinates = f"{coordinate[0]},{coordinate[1]}"
        execute_height = str(coordinate[2])
        waypoint_speed = placemark["properties"]["speed"]
        waypoint_heading_angle = placemark["properties"]["heading"]
        gimbal_angle = placemark["properties"]["gimbal_angle"]
        take_photo = placemark["properties"]["take_photo"]
    except IndexError as e:
        raise ValueError(str(e))

    placemark = ET.Element("Placemark")

    point = ET.SubElement(placemark, "Point")
    coordinates_elem = ET.SubElement(point, "coordinates")
    coordinates_elem.text = coordinates

    wpml_index = ET.SubElement(placemark, "wpml:index")
    wpml_index.text = str(index)

    execute_height_elem = ET.SubElement(placemark, "wpml:executeHeight")
    execute_height_elem.text = str(execute_height)

    waypoint_speed_elem = ET.SubElement(placemark, "wpml:waypointSpeed")
    waypoint_speed_elem.text = str(waypoint_speed)

    waypoint_heading_param = ET.SubElement(placemark, "wpml:waypointHeadingParam")
    wpml_waypoint_heading_mode = ET.SubElement(
        waypoint_heading_param, "wpml:waypointHeadingMode"
    )
    wpml_waypoint_heading_mode.text = "followWayline"

    wpml_waypoint_heading_angle = ET.SubElement(
        waypoint_heading_param, "wpml:waypointHeadingAngle"
    )
    wpml_waypoint_heading_angle.text = str(waypoint_heading_angle)
    wpml_waypoint_poi_point = ET.SubElement(
        waypoint_heading_param, "wpml:waypointPoiPoint"
    )
    wpml_waypoint_poi_point.text = "0.000000,0.000000,0.000000"
    wpml_waypoint_heading_angle_enable = ET.SubElement(
        waypoint_heading_param, "wpml:waypointHeadingAngleEnable"
    )
    wpml_waypoint_heading_angle_enable.text = "1"
    # wpml_waypoint_heading_angle_enable.text = "0"

    # wpml_waypoint_heading_path_mode = ET.SubElement(
    #     waypoint_heading_param, "wpml:waypointHeadingPathMode"
    # )
    # wpml_waypoint_heading_path_mode.text = "followBadArc"

    # wpml_waypoint_turn_param = ET.SubElement(placemark, "wpml:waypointTurnParam")
    # wpml_waypoint_turn_mode = ET.SubElement(
    #     wpml_waypoint_turn_param, "wpml:waypointTurnMode"
    # )
    # wpml_waypoint_turn_mode.text = "toPointAndStopWithContinuityCurvature"
    # wpml_waypoint_turn_damping_dist = ET.SubElement(
    #     wpml_waypoint_turn_param, "wpml:waypointTurnDampingDist"
    # )
    # wpml_waypoint_turn_damping_dist.text = "0"

    # use_straight_line = ET.SubElement(placemark, "wpml:useStraightLine")
    # use_straight_line.text = "0"

    action_group1 = ET.SubElement(placemark, "wpml:actionGroup")
    action_group1_id = ET.SubElement(action_group1, "wpml:actionGroupId")
    action_group1_id.text = "1"
    action_group1_start = ET.SubElement(action_group1, "wpml:actionGroupStartIndex")
    action_group1_start.text = str(index)  # Start index
    action_group1_end = ET.SubElement(action_group1, "wpml:actionGroupEndIndex")
    action_group1_end.text = str(index)  # End Index
    action_group1_mode = ET.SubElement(action_group1, "wpml:actionGroupMode")
    action_group1_mode.text = "parallel"
    action_group1_trigger = ET.SubElement(action_group1, "wpml:actionTrigger")
    action_group1_trigger_type = ET.SubElement(
        action_group1_trigger, "wpml:actionTriggerType"
    )
    action_group1_trigger_type.text = "reachPoint"

    if take_photo:
        # Take photo action
        take_photo_action(action_group1, "1")
    else:
        # Gimble rotate action
        gimble_rotate_action(action_group1, "1", str(gimbal_angle))

    action_group2 = ET.SubElement(placemark, "wpml:actionGroup")
    action_group2_id = ET.SubElement(action_group2, "wpml:actionGroupId")
    action_group2_id.text = "2"  # Not always 2
    action_group2_start = ET.SubElement(action_group2, "wpml:actionGroupStartIndex")
    action_group2_start.text = "0"
    action_group2_end = ET.SubElement(action_group2, "wpml:actionGroupEndIndex")
    action_group2_end.text = "1"
    action_group2_mode = ET.SubElement(action_group2, "wpml:actionGroupMode")
    action_group2_mode.text = "parallel"
    action_group2_trigger = ET.SubElement(action_group2, "wpml:actionTrigger")
    action_group2_trigger_type = ET.SubElement(
        action_group2_trigger, "wpml:actionTriggerType"
    )
    action_group2_trigger_type.text = "reachPoint"
    action2 = ET.SubElement(action_group2, "wpml:action")
    action2_id = ET.SubElement(action2, "wpml:actionId")
    action2_id.text = "2"  # Not always 2
    action2_actuator_func = ET.SubElement(action2, "wpml:actionActuatorFunc")
    action2_actuator_func.text = "gimbalEvenlyRotate"
    action2_actuator_func_param = ET.SubElement(action2, "wpml:actionActuatorFuncParam")
    gimbal_pitch_rotate_angle2 = ET.SubElement(
        action2_actuator_func_param, "wpml:gimbalPitchRotateAngle"
    )
    gimbal_pitch_rotate_angle2.text = str(gimbal_angle)
    payload_position_index2 = ET.SubElement(
        action2_actuator_func_param, "wpml:payloadPositionIndex"
    )
    payload_position_index2.text = "0"

    return placemark


def create_mission_config(finish_action_value, global_height):
    mission_config = ET.Element("wpml:missionConfig")

    fly_to_wayline_mode = ET.SubElement(mission_config, "wpml:flyToWaylineMode")
    fly_to_wayline_mode.text = "safely"

    finish_action = ET.SubElement(mission_config, "wpml:finishAction")
    finish_action.text = str(finish_action_value)

    exit_on_rc_lost = ET.SubElement(mission_config, "wpml:exitOnRCLost")
    exit_on_rc_lost.text = "executeLostAction"

    execute_rc_lost_action = ET.SubElement(mission_config, "wpml:executeRCLostAction")
    execute_rc_lost_action.text = "hover"

    global_transitional_speed = ET.SubElement(
        mission_config, "wpml:globalTransitionalSpeed"
    )
    global_transitional_speed.text = "2.5"

    global_rth_height = ET.SubElement(mission_config, "wpml:globalRTHHeight")
    global_rth_height.text = str(global_height)

    drone_info = ET.SubElement(mission_config, "wpml:droneInfo")
    drone_enum_value = ET.SubElement(drone_info, "wpml:droneEnumValue")
    drone_enum_value.text = "68"
    drone_sub_enum_value = ET.SubElement(drone_info, "wpml:droneSubEnumValue")
    drone_sub_enum_value.text = "0"

    return mission_config


def create_folder(placemarks):
    folder = ET.Element("Folder")

    template_id = ET.SubElement(folder, "wpml:templateId")
    template_id.text = "0"

    execute_height_mode = ET.SubElement(folder, "wpml:executeHeightMode")
    execute_height_mode.text = "relativeToStartPoint"

    wayline_id = ET.SubElement(folder, "wpml:waylineId")
    wayline_id.text = "0"

    distance = ET.SubElement(folder, "wpml:distance")
    distance.text = "0"

    duration = ET.SubElement(folder, "wpml:duration")
    duration.text = "0"

    global_waypoint_turn_mode = ET.SubElement(folder, "wpml:globalWaypointTurnMode")
    global_waypoint_turn_mode.text = "toPointAndStopWithDiscontinuityCurvature"

    straight_line = ET.SubElement(folder, "wpml:globalUseStraightLine")
    straight_line.text = "0"

    auto_flight_speed = ET.SubElement(folder, "wpml:autoFlightSpeed")
    auto_flight_speed.text = "2.5"

    for placemark in placemarks:
        placemark = create_placemark(placemark)
        folder.append(placemark)

    return folder


def create_kml(mission_config, folder):
    kml = ET.Element("kml")
    kml.set("xmlns", "http://www.opengis.net/kml/2.2")
    kml.set("xmlns:wpml", "http://www.dji.com/wpmz/1.0.2")

    document = ET.SubElement(kml, "Document")
    document.append(mission_config)
    document.append(folder)

    return kml


def create_xml(placemarks, finish_action, global_height, output_file_path="/tmp/"):
    mission_config = create_mission_config(finish_action, global_height)
    folder = create_folder(placemarks)
    kml = create_kml(mission_config, folder)

    tree = ET.ElementTree(kml)

    folder_name = "flight"
    os.makedirs(os.path.join(output_file_path, folder_name), exist_ok=True)
    waylines_path = os.path.join(output_file_path, folder_name, "waylines.wpml")

    tree.write(waylines_path, encoding="UTF-8", xml_declaration=True)
    output_file_name = create_zip_file(os.path.join(output_file_path, folder_name))
    return output_file_name


def create_wpml(
    placemark_geojson: Union[str, FeatureCollection, dict],
    output_file_path: str = "/tmp/",
):
    """
    Arguments:
        placemark_geojson: The placemark coordinates to be included in the flightplan mission
        output_file_path: The output file path for the wpml file
    Returns:
        wpml file.
    """

    # NOTE: finish action is set to "goHome" static.
    finish_action = "goHome"

    # global height is taken from the first point
    try:
        global_height = placemark_geojson["features"][0]["geometry"]["coordinates"][2]
    except IndexError:
        global_height = 100

    placemarks = placemark_geojson["features"]

    output_file = create_xml(placemarks, finish_action, global_height, output_file_path)

    return output_file


def main(args_list: list[str] | None = None):
    parser = argparse.ArgumentParser(
        description="Generate wpml file for drone missions."
    )
    parser.add_argument(
        "--placemark",
        required=True,
        type=str,
        help="The placemark geojson file to be included in the flightplan mission",
    )
    parser.add_argument(
        "--outfile",
        required=True,
        type=str,
        help="The output file path for the wpml file",
    )
    args = parser.parse_args(args_list)

    infile = open(args.placemark, "r")
    placemarks = infile.read()

    create_wpml(geojson.loads(placemarks), args.outfile)


if __name__ == "__main__":
    main()



import logging
import argparse

# Configure logging
logging.basicConfig(level=logging.INFO, format="%(message)s")
log = logging.getLogger(__name__)


def calculate_parameters(
    forward_overlap: float,
    side_overlap: float,
    agl: float,
    gsd: float = None,
    image_interval: int = 2,
):
    """
    Parameters
    ---------------------------------
    AGL(Altitude above ground level in meter ) = 115
    Forward overlap = 75
    Side overlap = 75

    ## Fixed Parameters
    Image interval = 2 sec
    Vertical FOV = 0.71
    Horizontal FOV = 1.26

    Forward Photo height = AGL * Vertical_FOV = 115*0.71 = 81.65
    Side Photo width = AGL * Horizontal_FOV = 115*1.26 = 144
    forward overlap distance =  forward photo height * forward overlap = 75 / 100 * 81.65 = 61.5
    side overlap distance = side photo width * side overlap = 75 / 100 * 144 = 108
    forward spacing =  forward photo height - forward overlap distance = 81.65 - 61.5 = 20.15
    side spacing = side photo width - side overlap distance = 144 - 108 = 36
    ground speed = forward spacing / image interval = 10

    """

    # Constants ( For DJI Mini 4 Pro)
    VERTICAL_FOV = 0.71
    HORIZONTAL_FOV = 1.26
    GSD_to_AGL_CONST = 29.7

    if gsd:
        agl = gsd * GSD_to_AGL_CONST

    # Calculations
    forward_photo_height = agl * VERTICAL_FOV
    side_photo_width = agl * HORIZONTAL_FOV
    forward_overlap_distance = forward_photo_height * forward_overlap / 100
    side_overlap_distance = side_photo_width * side_overlap / 100
    forward_spacing = forward_photo_height - forward_overlap_distance
    side_spacing = side_photo_width - side_overlap_distance
    ground_speed = forward_spacing / image_interval

    return {
        "forward_photo_height": round(forward_photo_height, 0),
        "side_photo_width": round(side_photo_width, 0),
        "forward_spacing": round(forward_spacing, 2),
        "side_spacing": round(side_spacing, 2),
        "ground_speed": round(ground_speed, 2),
        "altitude_above_ground_level": agl,
    }


def main():
    parser = argparse.ArgumentParser(
        description="Generate parameters for a drone which can be used for flight plans."
    )

    group = parser.add_mutually_exclusive_group(required=True)

    group.add_argument(
        "--altitude_above_ground_level",
        type=float,
        help="The flight altitude in meters.",
    )
    group.add_argument(
        "--gsd",
        type=float,
        help="The flight altitude in meters.",
    )

    parser.add_argument(
        "--forward_overlap",
        type=float,
        default=70.0,
        help="The forward overlap in percentage.",
    )
    parser.add_argument(
        "--side_overlap",
        type=float,
        default=70.0,
        help="The side overlap in percentage.",
    )
    parser.add_argument(
        "--image_interval",
        type=int,
        default=2,
        help="The time interval between two consecutive images in seconds.",
    )

    args = parser.parse_args()

    results = calculate_parameters(
        args.forward_overlap,
        args.side_overlap,
        args.altitude_above_ground_level,
        args.gsd,
        args.image_interval,
    )

    for key, value in results.items():
        log.info(f"{key}: {value}")

    return results


if __name__ == "__main__":
    main()
